unit Scanner;

// scanner - used to scan for tokens

interface

uses
	Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls,
	Utils, Emitter;

type

	TToken = ( tokLParen, tokRParen, tokEquals, tokPlus, tokMinus, tokTimes, tokDivide,
	tokGoto, tokIdent, tokNumber, tokShowKeyword, tokEOF, tokUnknown);



	TScanner = class
	private
		tokenCurrent: TToken;
		sValue: string;
		cPutBack: Char;
		iLineCount: Integer;
		sInputLine: String;
		sInputLineOld: String;
		iPosition: Integer;
		iValue: Integer;
		iGoodPosition: Integer;
		memo: TMemo;
		function GetChar: Char;
		function IsAlpha(c: Char): Boolean;
		procedure PutBack(c: Char);
		function IsCRLF(c: Char): Boolean;
	public
		constructor Create;
		destructor Destroy; override;
		procedure SetMemo(AMemo: TMemo);
		procedure Scan;
		procedure MatchScan(token: TToken);
		procedure Match(token: TToken);
		property Token: TToken read tokenCurrent;
		property StringValue: string read sValue;
		property NumberValue: integer read iValue;
		property CurrentLine: String read sInputLine;
		function CurrentLineCount: Integer;
		property CurrentPosition: Integer read iGoodPosition;
		function FullLine: String;
	end;

	ScanExcptn = class(Exception);
	ScanSyntaxErrorExcptn = class(ScanExcptn);



// ==========================================================================
implementation

const
	kcCR = #13;


{ TScanner }

// --------------------------------------------------------------------------
// class constructor
constructor TScanner.Create;
begin
	tokenCurrent := tokUnknown;
	sValue := '';
	cPutBack := #0;
	iLineCount := 0;
	sInputLine := '';
	sInputLineOld := '';
	iPosition := 0;
end;

// --------------------------------------------------------------------------
// class destructor
destructor TScanner.Destroy;
begin
	;
end;

// --------------------------------------------------------------------------
// set the input memo
procedure TScanner.SetMemo(AMemo: TMemo);
begin
	memo := AMemo;
end;

// --------------------------------------------------------------------------
// match the given token
procedure TScanner.Match(token: TToken);
begin
	if token <> tokenCurrent then
		raise ScanSyntaxErrorExcptn.create('Syntax error.');
end;

// --------------------------------------------------------------------------
// match the given token, then scan
procedure TScanner.MatchScan(token: TToken);
begin
	Match(token);
	Scan;
end;

// --------------------------------------------------------------------------
// scan for the next token
procedure TScanner.Scan;
var
	c: Char;
	c2: Char;
	lComment: Boolean;
begin

	iGoodPosition := length(sInputLine);

	tokenCurrent := tokUnknown;

	c := GetChar();

	lComment := true;

	while lComment do
	begin
		// skip whitespace
		while (c=' ') or (IsCRLF(c)) or (c=chr(9)) do
			c := GetChar();

		lComment := false;
		// skip comments like '// -----------'
		if  (c = '/') and (Trim(sInputLine) = '/') then
		begin
			c2 := GetChar();
			if c2 = '/' then
			begin
				lComment := true;
				while not IsCRLF(c) do
					c := GetChar();
			end
			else
				PutBack(c2);
		end;
	end;


	// eof?
	if iLineCount > memo.Lines.Count then
	begin
		tokenCurrent := tokEOF;
		exit;
	end;

	// single char tokens
	case c of
		'(':
			begin
				tokenCurrent := tokLParen;
				exit;
			end;
		')':
			begin
				tokenCurrent := tokRParen;
				exit;
			end;
		'*':
			begin
				tokenCurrent := tokTimes;
				exit;
			end;
		'/':
			begin
				tokenCurrent := tokDivide;
				exit;
			end;
		'+':
			begin
				tokenCurrent := tokPlus;
				exit;
			end;
		'-':
			begin
				tokenCurrent := tokMinus;
				exit;
			end;
		'=':
			begin
				tokenCurrent := tokEquals;
				exit;
			end;
		end;

	// double char tokens
	if c = '-' then
	begin
		c := GetChar();
		if c = '>' then
		begin
			tokenCurrent := tokGoto;
			exit;
		end
		else
			PutBack(c);
	end;

	// identifier or keyword
	if IsAlpha(c) then
	begin
		sValue := string(c);
		c := GetChar();
		while IsAlpha(c) do
		begin
			sValue := sValue + string(c);
			c := GetChar();
		end;
		tokenCurrent := tokIdent;
		PutBack(c);

		if lowercase(sValue) = 'show' then
			tokenCurrent := tokShowKeyword;
		// other keywords here...

		exit;
	end;

	(*
	// literal ie. in quotes
	if c = '''' then
	begin
		sValue := '';
		c := GetChar();
		while c <> '''' do
		begin
			if not IsCRLF(c) then
				sValue := sValue + string(c)
			else begin
				PutBack(c);
				exit; // incomplete
			end;
			c := GetChar();
		end;
		tokenCurrent := tokLiteral;
		exit;
	end;
	*)

	// number
	if (c >= '0') and (c <= '9') then
	begin
		iValue := ord(c) - ord('0');

		c := GetChar();
		while (c >= '0') and (c <= '9') do
		begin
			iValue := iValue * 10 + (ord(c) - ord('0'));
			c := GetChar();
		end;
		tokenCurrent := tokNumber;
		PutBack(c);
		exit;
	end;

end;

// --------------------------------------------------------------------------
// get the next char
function TScanner.GetChar: Char;
var
	c: Char;
begin

	result := #0; // default

	// eof?
	if iLineCount > memo.Lines.Count then
		exit;

	// advance to next
	Inc(iPosition);

	// end of line?
	if iPosition > length(memo.Lines[iLineCount]) then
	begin
		iPosition := 0;
		Inc(iLineCount);
		sInputLineOld := sInputLine;
		sInputLine := '';


		// insert artificial newline character
		//emit.Emit('<cr>');
		result := kcCR;
		exit;

	end;

	c := memo.Lines[iLineCount][iPosition];
	sInputLine := sInputLine + c;
	//emit.Emit('c=''' + c + '''; linecount = ' + IntToStr(iLineCount));

	result := c;
end;

// --------------------------------------------------------------------------
// put back a char
procedure TScanner.PutBack(c: Char);
begin
	cPutBack := c;

	if iPosition > 0 then
	begin
		sInputLine := StrLeft(sInputLine, length(sInputLine) - 1);
		Dec(iPosition);
	end
	else begin
		Dec(iLineCount);
		sInputLine := memo.Lines[iLineCount];
		iPosition := length(memo.Lines[iLineCount]) + 1;
	end;
end;

// --------------------------------------------------------------------------
// is the given character alphabetic?
function TScanner.IsAlpha(c: Char): Boolean;
begin
	result := ((c>= 'a') and (c<='z')) or ((c>= 'A') and (c<='Z'));
end;

// --------------------------------------------------------------------------
// is the given character a cr or lf?
function TScanner.IsCRLF(c: Char): Boolean;
begin
	result := (c=kcCR) or (c=chr(10));
end;

// --------------------------------------------------------------------------
// return the full line at the current position, do not scan
// - used in case of syntax error
function TScanner.FullLine: String;
begin
	result := memo.Lines[iLineCount];
end;

// --------------------------------------------------------------------------
// return the current line count
function TScanner.CurrentLineCount: Integer;
begin
	result := iLineCount + 1;
end;

end.
